# Autogenerated by Astropy-affiliated package astropy_helpers's setup.py on 2018-05-02 16:56:40
from __future__ import unicode_literals
import datetime

version = "3.0.1"
githash = "09bbe9583e70ed1804f1051a3aaca1640d08c6f8"


major = 3
minor = 0
bugfix = 1

release = True
timestamp = datetime.datetime(2018, 5, 2, 16, 56, 40)
debug = False

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
